import numpy as np
import pandas as pd
import torch
from sklearn.mixture import BayesianGaussianMixture


class DataTransformer():
    
    """
    Transformer class responsible for processing data to train the CTABGANSynthesizer model
    
    Variables:
    1) train_data -> input dataframe 
    2) categorical_list -> list of categorical columns
    3) mixed_dict -> dictionary of mixed columns
    4) n_clusters -> number of modes to fit bayesian gaussian mixture (bgm) model
    5) eps -> threshold for ignoring less prominent modes in the mixture model 
    6) ordering -> stores original ordering for modes of numeric columns
    7) output_info -> stores dimension and output activations of columns (i.e., tanh for numeric, softmax for categorical)
    8) output_dim -> stores the final column width of the transformed data
    9) components -> stores the valid modes used by numeric columns
    10) filter_arr -> stores valid indices of continuous component in mixed columns
    11) meta -> stores column information corresponding to different data types i.e., categorical/mixed/numerical


    Methods:
    1) __init__() -> initializes transformer object and computes meta information of columns
    2) get_metadata() -> builds an inventory of individual columns and stores their relevant properties
    3) fit() -> fits the required bgm models to process the input data
    4) transform() -> executes the transformation required to train the model
    5) inverse_transform() -> executes the reverse transformation on data generated from the model
    
    """    
    
    def __init__(self, train_data=pd.DataFrame, categorical_list=[], mixed_dict={}, n_clusters=10, eps=0.005):
        
        self.meta = None
        self.train_data = train_data
        self.categorical_columns= categorical_list
        self.mixed_columns= mixed_dict
        self.n_clusters = n_clusters
        self.eps = eps
        self.ordering = []
        self.output_info = []
        self.output_dim = 0
        self.components = []
        self.filter_arr = []
        self.meta = self.get_metadata()
        
    def get_metadata(self):
        
        meta = []
    
        for index in range(self.train_data.shape[1]):
            column = self.train_data.iloc[:,index]
            if index in self.categorical_columns:
                mapper = column.value_counts().index.tolist()
                meta.append({
                        "name": index,
                        "type": "categorical",
                        "size": len(mapper),
                        "i2s": mapper
                })
            elif index in self.mixed_columns.keys():
                meta.append({
                    "name": index,
                    "type": "mixed",
                    "min": column.min(),
                    "max": column.max(),
                    "modal": self.mixed_columns[index]
                })
            else:
                meta.append({
                    "name": index,
                    "type": "continuous",
                    "min": column.min(),
                    "max": column.max(),
                })            

        return meta

    def fit(self):
        
        data = self.train_data.values
        
        # stores the corresponding bgm models for processing numeric data
        model = []
        
        # iterating through column information
        for id_, info in enumerate(self.meta):
            if info['type'] == "continuous":
                # fitting bgm model  
                gm = BayesianGaussianMixture(
                    n_components=self.n_clusters, 
                    weight_concentration_prior_type='dirichlet_process',
                    weight_concentration_prior=0.001, # lower values result in lesser modes being active
                    max_iter=100,n_init=1, random_state=42)
                gm.fit(data[:, id_].reshape([-1, 1]))
                model.append(gm)
                # keeping only relevant modes that have higher weight than eps and are used to fit the data
                old_comp = gm.weights_ > self.eps
                mode_freq = (pd.Series(gm.predict(data[:, id_].reshape([-1, 1]))).value_counts().keys())
                comp = []
                for i in range(self.n_clusters):
                    if (i in (mode_freq)) & old_comp[i]:
                        comp.append(True)
                    else:
                        comp.append(False)
                self.components.append(comp) 
                self.output_info += [(1, 'tanh'), (np.sum(comp), 'softmax')]
                self.output_dim += 1 + np.sum(comp)
                
            elif info['type'] == "mixed":
                
                # in case of mixed columns, two bgm models are used
                gm1 = BayesianGaussianMixture(
                    n_components=self.n_clusters, 
                    weight_concentration_prior_type='dirichlet_process',
                    weight_concentration_prior=0.001, max_iter=100,
                    n_init=1,random_state=42)
                gm2 = BayesianGaussianMixture(
                    n_components=self.n_clusters,
                    weight_concentration_prior_type='dirichlet_process',
                    weight_concentration_prior=0.001, max_iter=100,
                    n_init=1,random_state=42)
                
                # first bgm model is fit to the entire data only for the purposes of obtaining a normalized value of any particular categorical mode
                gm1.fit(data[:, id_].reshape([-1, 1]))
                
                # main bgm model used to fit the continuous component and serves the same purpose as with purely numeric columns
                filter_arr = []
                for element in data[:, id_]:
                    if element not in info['modal']:
                        filter_arr.append(True)
                    else:
                        filter_arr.append(False)
                self.filter_arr.append(filter_arr)
                
                gm2.fit(data[:, id_][filter_arr].reshape([-1, 1]))
                
                model.append((gm1,gm2))
                
                # similarly keeping only relevant modes with higher weight than eps and are used to fit strictly continuous data 
                old_comp = gm2.weights_ > self.eps
                mode_freq = (pd.Series(gm2.predict(data[:, id_][filter_arr].reshape([-1, 1]))).value_counts().keys())  
                comp = []
                  
                for i in range(self.n_clusters):
                    if (i in (mode_freq)) & old_comp[i]:
                        comp.append(True)
                    else:
                        comp.append(False)

                self.components.append(comp)
                
                # modes of the categorical component are appended to modes produced by the main bgm model
                self.output_info += [(1, 'tanh'), (np.sum(comp) + len(info['modal']), 'softmax')]
                self.output_dim += 1 + np.sum(comp) + len(info['modal'])
            
            else:
                # in case of categorical columns, bgm model is ignored
                model.append(None)
                self.components.append(None)
                self.output_info += [(info['size'], 'softmax')]
                self.output_dim += info['size']
        
        self.model = model

    def transform(self, data):
        
        # stores the transformed values
        values = []

        # used for accessing filter_arr for transforming mixed columns
        mixed_counter = 0
        
        # iterating through column information
        for id_, info in enumerate(self.meta):
            current = data[:, id_]
            if info['type'] == "continuous":
                # mode-specific normalization occurs here
                current = current.reshape([-1, 1])
                # means and stds of the modes are obtained from the corresponding fitted bgm model
                means = self.model[id_].means_.reshape((1, self.n_clusters))
                stds = np.sqrt(self.model[id_].covariances_).reshape((1, self.n_clusters))
                # values are then normalized and stored for all modes
                features = np.empty(shape=(len(current),self.n_clusters))
                # note 4 is a multiplier to ensure values lie between -1 to 1 but this is not always guaranteed
                features = (current - means) / (4 * stds) 

                # number of distict modes
                n_opts = sum(self.components[id_])                
                # storing the mode for each data point by sampling from the probability mass distribution across all modes based on fitted bgm model 
                opt_sel = np.zeros(len(data), dtype='int')
                probs = self.model[id_].predict_proba(current.reshape([-1, 1]))
                probs = probs[:, self.components[id_]]
                for i in range(len(data)):
                    pp = probs[i] + 1e-6
                    pp = pp / sum(pp)
                    opt_sel[i] = np.random.choice(np.arange(n_opts), p=pp)
                
                # creating a one-hot-encoding for the corresponding selected modes
                probs_onehot = np.zeros_like(probs)
                probs_onehot[np.arange(len(probs)), opt_sel] = 1

                # obtaining the normalized values based on the appropriately selected mode and clipping to ensure values are within (-1,1)
                idx = np.arange((len(features)))
                features = features[:, self.components[id_]]
                features = features[idx, opt_sel].reshape([-1, 1])
                features = np.clip(features, -.99, .99) 
                
                # re-ordering the one-hot-encoding of modes in descending order as per their frequency of being selected
                re_ordered_phot = np.zeros_like(probs_onehot)  
                col_sums = probs_onehot.sum(axis=0)
                n = probs_onehot.shape[1]
                largest_indices = np.argsort(-1*col_sums)[:n]
                for id,val in enumerate(largest_indices):
                    re_ordered_phot[:,id] = probs_onehot[:,val]
                
                # storing the original ordering for invoking inverse transform
                self.ordering.append(largest_indices)
                
                # storing transformed numeric column represented as normalized values and corresponding modes 
                values += [features, re_ordered_phot]
                  
            elif info['type'] == "mixed":
                
                # means and standard deviation of modes obtained from the first fitted bgm model
                means_0 = self.model[id_][0].means_.reshape([-1])
                stds_0 = np.sqrt(self.model[id_][0].covariances_).reshape([-1])

                # list to store relevant bgm modes for categorical components
                zero_std_list = []
                
                # means and stds needed to normalize relevant categorical components
                means_needed = []
                stds_needed = []

                # obtaining the closest bgm mode to the categorical component
                for mode in info['modal']:
                    # skipped for mode representing missing values
                    if mode!=-9999999:
                        dist = []
                        for idx,val in enumerate(list(means_0.flatten())):
                            dist.append(abs(mode-val))
                        index_min = np.argmin(np.array(dist))
                        zero_std_list.append(index_min)
                    else: continue

                
                # stores the appropriate normalized value of categorical modes
                mode_vals = []
                
                # based on the means and stds of the chosen modes for categorical components, their respective values are similarly normalized
                for idx in zero_std_list:
                    means_needed.append(means_0[idx])
                    stds_needed.append(stds_0[idx])
               
                for i,j,k in zip(info['modal'],means_needed,stds_needed):
                    this_val  = np.clip(((i - j) / (4*k)), -.99, .99) 
                    mode_vals.append(this_val)
                
                # for categorical modes representing missing values, the normalized value associated is simply 0
                if -9999999 in info["modal"]:
                    mode_vals.append(0)
                
                # transforming continuous component of mixed columns similar to purely numeric columns using second fitted bgm model
                current = current.reshape([-1, 1])
                filter_arr = self.filter_arr[mixed_counter]
                current = current[filter_arr]
    
                means = self.model[id_][1].means_.reshape((1, self.n_clusters))
                stds = np.sqrt(self.model[id_][1].covariances_).reshape((1, self.n_clusters))
                
                features = np.empty(shape=(len(current),self.n_clusters))
                features = (current - means) / (4 * stds)
                
                n_opts = sum(self.components[id_]) 
                probs = self.model[id_][1].predict_proba(current.reshape([-1, 1]))
                probs = probs[:, self.components[id_]]
                
                opt_sel = np.zeros(len(current), dtype='int')
                for i in range(len(current)):
                    pp = probs[i] + 1e-6
                    pp = pp / sum(pp)
                    opt_sel[i] = np.random.choice(np.arange(n_opts), p=pp)
                
                idx = np.arange((len(features)))
                features = features[:, self.components[id_]]
                features = features[idx, opt_sel].reshape([-1, 1])
                features = np.clip(features, -.99, .99)
                
                probs_onehot = np.zeros_like(probs)
                probs_onehot[np.arange(len(probs)), opt_sel] = 1
                
                # additional modes are appended to represent categorical component
                extra_bits = np.zeros([len(current), len(info['modal'])])
                temp_probs_onehot = np.concatenate([extra_bits,probs_onehot], axis = 1)
                
                # storing the final normalized value and one-hot-encoding of selected modes
                final = np.zeros([len(data), 1 + probs_onehot.shape[1] + len(info['modal'])])

                # iterates through only the continuous component
                features_curser = 0

                for idx, val in enumerate(data[:, id_]):
                    
                    if val in info['modal']:
                        # dealing with the modes of categorical component
                        category_ = list(map(info['modal'].index, [val]))[0]
                        final[idx, 0] = mode_vals[category_]
                        final[idx, (category_+1)] = 1
                    
                    else:
                        # dealing with the modes of continuous component
                        final[idx, 0] = features[features_curser]
                        final[idx, (1+len(info['modal'])):] = temp_probs_onehot[features_curser][len(info['modal']):]
                        features_curser = features_curser + 1

                # re-ordering the one-hot-encoding of modes in descending order as per their frequency of being selected
                just_onehot = final[:,1:]
                re_ordered_jhot= np.zeros_like(just_onehot)
                n = just_onehot.shape[1]
                col_sums = just_onehot.sum(axis=0)
                largest_indices = np.argsort(-1*col_sums)[:n]
                
                for id,val in enumerate(largest_indices):
                      re_ordered_jhot[:,id] = just_onehot[:,val]
                
                final_features = final[:,0].reshape([-1, 1])
                
                # storing the original ordering for invoking inverse transform
                self.ordering.append(largest_indices)
                
                values += [final_features, re_ordered_jhot]
                
                mixed_counter = mixed_counter + 1
    
            else:
                # for categorical columns, standard one-hot-encoding is applied where categories are in descending order of frequency by default
                self.ordering.append(None)
                col_t = np.zeros([len(data), info['size']])
                idx = list(map(info['i2s'].index, current))
                col_t[np.arange(len(data)), idx] = 1
                values.append(col_t)
                
        return np.concatenate(values, axis=1)

    def inverse_transform(self, data):
        
        # stores the final inverse transformed generated data 
        data_t = np.zeros([len(data), len(self.meta)])
        
        # used to iterate through the columns of the raw generated data
        st = 0

        # iterating through original column information
        for id_, info in enumerate(self.meta):
            if info['type'] == "continuous":
                
                # obtaining the generated normalized values and clipping for stability
                u = data[:, st]
                u = np.clip(u, -1, 1)
                
                # obtaining the one-hot-encoding of the modes representing the normalized values
                v = data[:, st + 1:st + 1 + np.sum(self.components[id_])]
                
                # re-ordering the modes as per their original ordering
                order = self.ordering[id_] 
                v_re_ordered = np.zeros_like(v)
                for id,val in enumerate(order):
                    v_re_ordered[:,val] = v[:,id]
                v = v_re_ordered

                # ensuring un-used modes are represented with -100 such that they can be ignored when computing argmax
                v_t = np.ones((data.shape[0], self.n_clusters)) * -100
                v_t[:, self.components[id_]] = v
                v = v_t
                
                # obtaining approriate means and stds as per the appropriately selected mode for each data point based on fitted bgm model
                means = self.model[id_].means_.reshape([-1])
                stds = np.sqrt(self.model[id_].covariances_).reshape([-1])
                p_argmax = np.argmax(v, axis=1)
                std_t = stds[p_argmax]
                mean_t = means[p_argmax]
                
                # executing the inverse transformation 
                tmp = u * 4 * std_t + mean_t
                
                data_t[:, id_] = tmp
                
                # moving to the next set of columns in the raw generated data in correspondance to original column information
                st += 1 + np.sum(self.components[id_])
                
            elif info['type'] == "mixed":
                
                # obtaining the generated normalized values and corresponding modes
                u = data[:, st]
                u = np.clip(u, -1, 1)
                full_v = data[:,(st+1):(st+1)+len(info['modal'])+np.sum(self.components[id_])]
                
                # re-ordering the modes as per their original ordering
                order = self.ordering[id_]
                full_v_re_ordered = np.zeros_like(full_v)
                for id,val in enumerate(order):
                    full_v_re_ordered[:,val] = full_v[:,id]
                full_v = full_v_re_ordered                
                
                # modes of categorical component
                mixed_v = full_v[:,:len(info['modal'])]
                
                # modes of continuous component
                v = full_v[:,-np.sum(self.components[id_]):]

                # similarly ensuring un-used modes are represented with -100 to be ignored while computing argmax
                v_t = np.ones((data.shape[0], self.n_clusters)) * -100
                v_t[:, self.components[id_]] = v
                v = np.concatenate([mixed_v,v_t], axis=1)       
                p_argmax = np.argmax(v, axis=1)

                # obtaining the means and stds of the continuous component using second fitted bgm model
                means = self.model[id_][1].means_.reshape([-1]) 
                stds = np.sqrt(self.model[id_][1].covariances_).reshape([-1]) 

                # used to store the inverse-transformed data points
                result = np.zeros_like(u)

                for idx in range(len(data)):
                    # in case of categorical mode being selected, the mode value itself is simply assigned 
                    if p_argmax[idx] < len(info['modal']):
                        argmax_value = p_argmax[idx]
                        result[idx] = float(list(map(info['modal'].__getitem__, [argmax_value]))[0])
                    else:
                        # in case of continuous mode being selected, similar inverse-transform for purely numeric values is applied
                        std_t = stds[(p_argmax[idx]-len(info['modal']))]
                        mean_t = means[(p_argmax[idx]-len(info['modal']))]
                        result[idx] = u[idx] * 4 * std_t + mean_t
            
                data_t[:, id_] = result

                st += 1 + np.sum(self.components[id_]) + len(info['modal'])
                
            else:
                # reversing one hot encoding back to label encoding for categorical columns 
                current = data[:, st:st + info['size']]
                idx = np.argmax(current, axis=1)
                data_t[:, id_] = list(map(info['i2s'].__getitem__, idx))
                st += info['size']
        return data_t

class ImageTransformer():

    """
    Transformer responsible for translating data rows to images and vice versa

    Variables:
    1) side -> height/width of the image

    Methods:
    1) __init__() -> initializes image transformer object with given input
    2) transform() -> converts tabular data records into square image format
    3) inverse_transform() -> converts square images into tabular format

    """
    
    def __init__(self, side):
    
        self.height = side
            
    def transform(self, data):
        
        if self.height * self.height > len(data[0]):
            # tabular data records are padded with 0 to conform to square shaped images
            padding = torch.zeros((len(data), self.height * self.height - len(data[0]))).to(data.device)
            data = torch.cat([data, padding], axis=1)

        return data.view(-1, 1, self.height, self.height)

    def inverse_transform(self, data):
        
        data = data.view(-1, self.height * self.height)

        return data





from . import Tuner
from gentab.generators import CTABGANPlus
from gentab.evaluators import Evaluator
from gentab.utils import console, SPINNER, REFRESH

import optuna


class CTABGANPlusTuner(Tuner):
    def __init__(
        self,
        evaluator: Evaluator,
        trials: int,
        *args,
        min_epochs: int = 300,
        max_epochs: int = 800,
        min_batch: int = 512,
        max_batch: int = 16384,
        max_tries_per_batch: int = 8192,
        timeout: int = None,
        **kwargs,
    ) -> None:
        super().__init__(
            evaluator,
            trials,
            min_epochs=min_epochs,
            max_epochs=max_epochs,
            min_batch=min_batch,
            max_batch=max_batch,
            max_tries_per_batch=max_tries_per_batch,
            timeout=timeout,
        )

    def objective(self, trial: optuna.trial.Trial) -> float:
        test_ratio = trial.suggest_float("test_ratio", 0.1, 0.3)
        epochs = trial.suggest_int("epochs", self.min_epochs, self.max_epochs)
        batch_size = trial.suggest_categorical("batch_size", self.batch_sizes)
        class_dim = trial.suggest_categorical(
            "class_dim",
            [
                (128, 256, 256, 128),
                (256, 256, 256, 256),
                (512, 512, 512, 512),
            ],
        )
        random_dim = trial.suggest_int("random_dim", 50, 200)
        num_channels = trial.suggest_categorical("num_channels", [16, 32, 64, 128])
        l2scale = trial.suggest_float("l2scale", 1e-6, 1e-3, log=True)

        self.generator = CTABGANPlus(
            self.dataset,
            test_ratio=test_ratio,
            epochs=epochs,
            class_dim=class_dim,
            num_channels=num_channels,
            random_dim=random_dim,
            l2scale=l2scale,
            batch_size=batch_size,
            max_tries_per_batch=self.max_tries_per_batch,
        )
        self.generator.generate()

        self.store_data(trial)

        acc, mcc = self.evaluator.evaluate(validation=True)

        return mcc



from gentab.evaluators import KNN, LightGBM, XGBoost, MLP
from gentab.generators import (
    ROS,
    SMOTE,
    ADASYN,
    TVAE,
    CTGAN,
    GaussianCopula,
    CopulaGAN,
    CTABGAN,
    CTABGANPlus,
    AutoDiffusion,
    ForestDiffusion,
    Tabula,
    GReaT,
)
from gentab.data import Config, Dataset
from gentab.utils import console

import matplotlib.pyplot as plt
import matplotlib as mpl
import matplotlib.lines as lines
from mpl_toolkits.axes_grid1 import ImageGrid


def preproc_playnet(dataset):
    dataset.reduce_size(
        {
            "left_attack": 0.97,
            "right_attack": 0.97,
            "right_transition": 0.9,
            "left_transition": 0.9,
            "time_out": 0.8,
            "left_penal": 0.5,
            "right_penal": 0.5,
        }
    )
    dataset.merge_classes(
        {
            "attack": ["left_attack", "right_attack"],
            "transition": ["left_transition", "right_transition"],
            "penalty": ["left_penal", "right_penal"],
        }
    )
    dataset.reduce_mem()

    return dataset


def preproc_adult(dataset):
    dataset.merge_classes({"<=50K": ["<=50K."], ">50K": [">50K."]})

    return dataset


def preproc_car(dataset):
    return dataset


configs = [
    ("configs/car_evaluation_cr.json", preproc_car, "Car Evaluation"),
    ("configs/playnet_cr.json", preproc_playnet, "PlayNet"),
    ("configs/adult_cr.json", preproc_adult, "Adult"),
]

gens = [
    (TVAE, "TVAE"),
    (CTGAN, "CTGAN"),
    (GaussianCopula, "Gaussian Copula"),
    (CopulaGAN, "Copula GAN"),
    (CTABGAN, "CTAB-GAN"),
    (CTABGANPlus, "CTAB-GAN+"),
    (AutoDiffusion, "AutoDiffusion"),
    (ForestDiffusion, "ForestDiffusion"),
    (GReaT, "GReaT"),
    (Tabula, "Tabula"),
]

for c in configs:
    config = Config(c[0])

    dataset = c[1](Dataset(config))
    console.print(dataset.class_counts(), dataset.row_count())

    corrs = []
    for g in gens:
        generator = g[0](dataset)
        try:
            generator.load_from_disk()
        except FileNotFoundError:
            corrs.append(None)
            continue

        corrs.append(dataset.get_pearson_correlation())

    max_corr = max(map(lambda x: x.values.max() if x is not None else 0, corrs))

    fig = plt.figure(figsize=(20, 10))

    grid = ImageGrid(
        fig,
        111,  # similar to subplot(111)
        nrows_ncols=(2, len(gens) // 2),
        axes_pad=0.8,
        direction="row",
        share_all=True,
        cbar_mode="single",
        cbar_location="right",
    )

    axs, ims = [], []
    i = 0
    for corr in corrs:
        ax = grid.axes_all[i]
        ax.set_title(gens[i][1])
        ax.spines["top"].set_visible(False)
        ax.spines["right"].set_visible(False)
        ax.spines["bottom"].set_visible(False)
        ax.spines["left"].set_visible(False)
        ax.get_xaxis().set_ticks([])
        ax.get_yaxis().set_ticks([])

        if corr is not None:
            axs.append(ax)
            ims.append(
                ax.imshow(
                    corr,
                    cmap="Greens",
                    interpolation="nearest",
                    vmin=0.0,
                    vmax=max_corr,
                )
            )
        else:
            axs.append(ax)
            # Iterating over the grid returns the Axes.
            l1 = lines.Line2D([0, 1], [0, 1], transform=fig.transFigure, figure=fig)
            l2 = lines.Line2D([0, 1], [1, 0], transform=fig.transFigure, figure=fig)
            # fig.lines.extend([l1, l2])
            ax.plot(
                [0, dataset.num_features()], [0, dataset.num_features()], color="red"
            )  # Diagonal from bottom-left to top-right

            ims.append(
                ax.plot(
                    [0, dataset.num_features()],
                    [dataset.num_features(), 0],
                    color="red",
                )  # Diagonal from top-left to bottom-right
            )

        i += 1

    cb = fig.colorbar(ims[0], cax=grid.cbar_axes[0], orientation="vertical")
    cb.outline.set_visible(False)

    plt.savefig(
        "figures/Correlation" + c[2] + ".pdf", format="pdf", bbox_inches="tight"
    )



from . import Generator
from .autodiff.process_GQ import DataFrameParser, convert_to_table
from .autodiff.autoencoder import train_autoencoder
from .autodiff.TabDDPMdiff import train_diffusion
from .autodiff.diffusion import Euler_Maruyama_sampling
from gentab.utils import console, PROG_COLUMNS, DEVICE

import pandas as pd


class AutoDiffusion(Generator):
    def __init__(
        self,
        dataset,
        threshold=0.01,
        device=DEVICE,
        max_tries_per_batch=8192,
        n_epochs=10000,
        eps=1e-5,
        weight_decay=1e-6,
        maximum_learning_rate=1e-2,
        lr=2e-4,
        hidden_size=250,
        num_layers=3,
        batch_size=100,
        diff_n_epochs=10000,
        hidden_dims=(256, 512, 1024, 512, 256),
        sigma=20,
        num_batches_per_epoch=50,
        T=100,
    ) -> None:
        super().__init__(dataset, batch_size, max_tries_per_batch)
        self.threshold = threshold
        # Auto-encoder hyper-parameters
        self.device = device
        self.n_epochs = n_epochs
        self.eps = eps
        self.weight_decay = weight_decay
        self.maximum_learning_rate = maximum_learning_rate
        self.lr = lr
        self.hidden_size = hidden_size
        self.num_layers = num_layers
        # diffusion hyper-parameters
        self.diff_n_epochs = diff_n_epochs
        self.hidden_dims = hidden_dims
        self.sigma = sigma
        self.num_batches_per_epoch = num_batches_per_epoch
        self.T = T
        self.data = self.dataset.get_single_df()

    def preprocess(self) -> None:
        super().preprocess()

    def train(self) -> None:
        # Setup progress
        self.p.columns = PROG_COLUMNS
        self.p.update(
            self.gen_task,
            total=self.n_epochs,
            description="Training {} AE...".format(self),
        )
        dm_task = self.p.add_task(
            total=self.diff_n_epochs,
            description="Training {} DM...".format(self),
        )

        self.ds = train_autoencoder(
            self.data,
            self.hidden_size,
            self.num_layers,
            self.lr,
            self.weight_decay,
            self.n_epochs,
            self.batch_size,
            self.threshold,
            self.p,
            self.gen_task,
        )
        self.latent_features = self.ds[1].detach()
        self.converted_table_dim = self.latent_features.shape[1]

        self.score = train_diffusion(
            self.latent_features,
            self.T,
            self.eps,
            self.sigma,
            self.lr,
            self.num_batches_per_epoch,
            self.maximum_learning_rate,
            self.weight_decay,
            self.diff_n_epochs,
            self.batch_size,
            self.device,
            self.p,
            dm_task,
        )

    def sample(self) -> pd.DataFrame:
        T = self.T
        N = self.latent_features.shape[0]
        P = self.latent_features.shape[1]

        sample = Euler_Maruyama_sampling(self.score, T, N, P, self.device)
        gen_output = self.ds[0](sample, self.ds[2], self.ds[3])

        return convert_to_table(self.data, gen_output, self.threshold)



import time
from gentab import SEED
from gentab.utils import console, ProgressBar, Timer

import pandas as pd


class Generator:
    def __init__(
        self, dataset, batch_size=1000, max_tries_per_batch=1000, seed=SEED
    ) -> None:
        self.batch_size = batch_size
        self.max_tries_per_batch = max_tries_per_batch
        # Reset generated dataframes
        self.dataset = dataset
        self.config = dataset.config
        self.dataset.X_gen = None
        self.dataset.y_gen = None
        # Get class counts
        self.orig_counts = dataset.class_counts()
        # Get max, subtract it to the rest to get target samples
        self.counts = self.orig_counts.max() - self.orig_counts
        # Filter the maximum class no need to generate for it
        self.counts = self.counts[self.counts > 0]
        self.target_counts = self.counts
        # For reproducibility
        self.seed = seed
        self.timer = Timer()

    def __str__(self) -> str:
        return self.__class__.__name__

    def preprocess(self) -> None:
        pass

    def train(self) -> None:
        pass

    def sample(self) -> pd.DataFrame:
        pass

    # def save_to_disk(self, tuner: str = "") -> None:
    #     self.dataset.save_to_disk(self, tuner)

    def load_from_disk(self, tuner: str = "") -> None:
        self.dataset.load_from_disk(self)

    def resample(self, n_samples, append) -> None:
        n_samples = n_samples.copy()

        if append:
            data_gen = self.dataset.get_single_df()
        else:
            data_gen = pd.DataFrame()

        total_samples = sum(n_samples.values())

        with ProgressBar().progress as p:
            self.p = p
            gen_task = p.add_task(
                "Generating with {}...".format(self), total=total_samples
            )
            for i in range(self.max_tries_per_batch):
                gen = self.sample()

                for cls, cnt in n_samples.items():
                    if cnt > 0:
                        filtered = gen[gen[self.dataset.config["y_label"]] == cls]

                        count = len(filtered.index)
                        if count > cnt:
                            n_samples[cls] = 0
                            filtered = filtered.sample(n=cnt)
                        else:
                            n_samples[cls] = cnt - count

                        data_gen = pd.concat(
                            [data_gen, filtered], ignore_index=True, sort=False
                        )

                missing_samples = sum(n_samples.values())

                p.update(gen_task, completed=total_samples - missing_samples)

                if missing_samples == 0:
                    break
                elif i == self.max_tries_per_batch - 1:
                    raise RuntimeError(
                        "Maximum number of tries reached, model probably did not"
                        " converge."
                    )

        self.dataset.set_split_result(data_gen)

    def balance(self) -> None:
        data_gen = self.dataset.get_single_df()

        total_samples = self.target_counts.sum()
        self.counts = self.target_counts

        with ProgressBar().progress as p:
            self.p = p
            gen_task = p.add_task(
                "Generating with {}...".format(self), total=total_samples
            )

            for i in range(self.max_tries_per_batch):
                gen = self.sample()

                for cls, cnt in self.counts.items():
                    if cnt > 0:
                        filtered = gen[gen[self.dataset.config["y_label"]] == cls]

                        count = len(filtered.index)
                        if count > cnt:
                            self.counts[cls] = 0
                            filtered = filtered.sample(n=cnt)
                        else:
                            self.counts[cls] = cnt - count

                        data_gen = pd.concat(
                            [data_gen, filtered], ignore_index=True, sort=False
                        )

                missing_samples = self.counts.sum()

                p.update(gen_task, completed=total_samples - missing_samples)

                if self.counts.max() < 1:
                    break
                elif i == self.max_tries_per_batch - 1:
                    raise RuntimeError(
                        "Maximum number of tries reached, model probably did not"
                        " converge."
                    )

        self.dataset.set_split_result(data_gen)

    def generate(self, n_samples=None, append=True, train=True) -> None:
        if train:
            with ProgressBar(indeterminate=True).progress as p:
                self.p = p

                self.timer.start()

                self.gen_task = self.p.add_task(
                    "Preprocessing with {}...".format(self.dataset), total=None
                )
                self.preprocess()

                self.p.update(
                    self.gen_task, description="Training with {}...".format(self)
                )
                self.train()

                console.print("✅ Training complete with {}...".format(self))

                self.timer.stop()
                self.timer.elapsed()

        self.timer.start()

        # If progress does not look better indent this
        if n_samples is None:
            self.balance()
        else:
            self.resample(n_samples, append)

        console.print("✅ Generation complete with {}...".format(self))

        self.timer.stop()
        self.timer.elapsed()

    def benchmark(self):
        self.generate()

        return self.timer.history



from gentab.evaluators import KNN, LightGBM, XGBoost, MLP
from gentab.generators import (
    ROS,
    SMOTE,
    ADASYN,
    TVAE,
    CTGAN,
    GaussianCopula,
    CopulaGAN,
    CTABGAN,
    CTABGANPlus,
    AutoDiffusion,
    ForestDiffusion,
    Tabula,
    GReaT,
)
from gentab.data import Config, Dataset
from gentab.utils import console

import matplotlib as mpl
import matplotlib.gridspec as gridspec
import matplotlib.pyplot as plt
from matplotlib.patches import Rectangle, Circle, Ellipse, Arc
import numpy as np
from scipy.stats import kde


def preproc_playnet(dataset):
    dataset.reduce_size(
        {
            "left_attack": 0.97,
            "right_attack": 0.97,
            "right_transition": 0.9,
            "left_transition": 0.9,
            "time_out": 0.8,
            "left_penal": 0.5,
            "right_penal": 0.5,
        }
    )
    dataset.merge_classes(
        {
            "attack": ["left_attack", "right_attack"],
            "transition": ["left_transition", "right_transition"],
            "penalty": ["left_penal", "right_penal"],
        }
    )
    dataset.reduce_mem()

    return dataset


gens = [
    (None, "Baseline"),
    (TVAE, "TVAE"),
    (CTGAN, "CTGAN"),
    (GaussianCopula, "Gaussian Copula"),
    (CopulaGAN, "Copula GAN"),
    (CTABGAN, "CTAB-GAN"),
    (CTABGANPlus, "CTAB-GAN+"),
    (AutoDiffusion, "AutoDiffusion"),
    (ForestDiffusion, "ForestDiffusion"),
    (GReaT, "GReaT"),
    (Tabula, "Tabula"),
]

config = Config("configs/playnet_cr.json")

dataset = preproc_playnet(Dataset(config))

gs = gridspec.GridSpec(len(gens), dataset.num_classes())
fig = plt.figure(figsize=(40, 30))
fig.subplots_adjust(wspace=0.25, hspace=0.125)

axs, ims = [], []
maxcnt = 0
i = 0
for g in gens:
    if g[0] is not None:
        generator = g[0](dataset)
        generator.load_from_disk()

    j = 0
    for c in dataset.class_names():
        ax = fig.add_subplot(gs[i, j])

        # Set labels
        if i == 0:
            ax.set_title("timeout" if c == "time_out" else c, fontsize=19)
        if j == 0:
            if g[0] is not None:
                ax.set_ylabel(g[1], fontsize=18)
            else:
                ax.set_ylabel(g[1], fontsize=18)

        # Remove chart borders and ticks
        ax.spines["top"].set_visible(False)
        ax.spines["right"].set_visible(False)
        ax.spines["bottom"].set_visible(False)
        ax.spines["left"].set_visible(False)
        ax.get_xaxis().set_ticks([])
        ax.get_yaxis().set_ticks([])

        # Get positions
        if g[0] is not None:
            row = generator.dataset.get_gen_class_rows(c)
        else:
            row = dataset.get_class_rows(c)

        tolerance = 0.05
        xpoints = row.filter(regex="^#x").values.flatten()
        ypoints = row.filter(regex="^#y").values.flatten()
        xpoints[np.isclose(xpoints, 0, atol=tolerance)] = 0
        ypoints[np.isclose(ypoints, 0, atol=tolerance)] = 0
        ids = ~(np.array(xpoints == 0) & np.array(ypoints == 0))
        xpoints = xpoints[ids]
        ypoints = ypoints[ids]
        # First array horiz. coords., second vertical
        # Player position heatmap
        # ax.hist2d(
        #     xpoints,
        #     ypoints,
        #     # bins=[np.arange(0, 1.0, 0.08), np.arange(0.0, 1.0, 0.08)],
        #     cmap="Greens",
        # )
        # Evaluate a gaussian kde on a regular grid of nbins x nbins over data extents
        nbins = 20
        k = kde.gaussian_kde([xpoints, ypoints])
        xi, yi = np.mgrid[
            0 : 1 : nbins * 1j,
            0 : 1 : nbins * 1j,
        ]
        zi = k(np.vstack([xi.flatten(), yi.flatten()]))
        maxcnt = zi.max() if zi.max() > maxcnt else maxcnt
        # Make the plot
        ax.pcolormesh(xi, yi, zi.reshape(xi.shape), shading="auto", cmap="Greens")
        # Middle line
        ax.plot([0.5, 0.5], [0.0, 1.0], color="black")
        # Center circle
        ax.add_patch(Ellipse((0.5, 0.5), 0.2, 0.7, facecolor="none", ec="k", lw=2))
        # Areas
        ax.add_patch(
            Arc(
                (0.0, 0.5),
                0.2,
                0.9,
                angle=180,
                theta1=90,
                theta2=270,
                facecolor="none",
                ec="k",
                lw=2,
            )
        )
        ax.add_patch(
            Arc(
                (1.0, 0.5),
                0.2,
                0.9,
                theta1=90,
                theta2=270,
                facecolor="none",
                ec="k",
                lw=2,
            )
        )
        # Field
        ims.append(
            ax.add_patch(Rectangle((0, 0), 1, 1, facecolor="none", ec="k", lw=2))
        )

        axs.append(ax)

        j += 1

    i += 1

cmap = mpl.cm.Greens
norm = mpl.colors.Normalize(vmin=0, vmax=round(maxcnt))
cb = fig.colorbar(
    mpl.cm.ScalarMappable(norm=norm, cmap=cmap),
    ax=axs,
    orientation="vertical",
    fraction=0.1,
    aspect=60,
    pad=0.025,
)
cb.ax.tick_params(labelsize=20)
cb.outline.set_visible(False)

plt.savefig("figures/HeatmapPlaynet.pdf", format="pdf", bbox_inches="tight")



from gentab.evaluators import KNN, LightGBM, XGBoost, MLP
from gentab.generators import (
    ROS,
    SMOTE,
    ADASYN,
    TVAE,
    CTGAN,
    GaussianCopula,
    CopulaGAN,
    CTABGAN,
    CTABGANPlus,
    AutoDiffusion,
    ForestDiffusion,
    Tabula,
    GReaT,
)
from gentab.data import Config, Dataset
from gentab.utils import console

import matplotlib.gridspec as gridspec
import matplotlib.pyplot as plt
from matplotlib.patches import Rectangle, Circle, Ellipse, Arc
import numpy as np


def preproc_playnet(dataset):
    dataset.reduce_size(
        {
            "left_attack": 0.97,
            "right_attack": 0.97,
            "right_transition": 0.9,
            "left_transition": 0.9,
            "time_out": 0.8,
            "left_penal": 0.5,
            "right_penal": 0.5,
        }
    )
    dataset.merge_classes(
        {
            "attack": ["left_attack", "right_attack"],
            "transition": ["left_transition", "right_transition"],
            "penalty": ["left_penal", "right_penal"],
        }
    )
    dataset.reduce_mem()

    return dataset


gens = [
    (None, "Baseline"),
    (TVAE, "TVAE"),
    (CTGAN, "CTGAN"),
    (GaussianCopula, "Gaussian Copula"),
    (CopulaGAN, "Copula GAN"),
    (CTABGAN, "CTAB-GAN"),
    (CTABGANPlus, "CTAB-GAN+"),
    (AutoDiffusion, "AutoDiffusion"),
    (ForestDiffusion, "ForestDiffusion"),
    (GReaT, "GReaT"),
    (Tabula, "Tabula"),
]

config = Config("configs/playnet_cr.json")

dataset = preproc_playnet(Dataset(config))

gs = gridspec.GridSpec(len(gens), dataset.num_classes())
fig = plt.figure(figsize=(40, 30))
fig.subplots_adjust(wspace=0.25, hspace=0.125)

axs, ims = [], []

i = 0
for g in gens:
    if g[0] is not None:
        generator = g[0](dataset)
        generator.load_from_disk()

    j = 0
    for c in dataset.class_names():
        ax = fig.add_subplot(gs[i, j])

        # Set labels
        if i == 0:
            ax.set_title("timeout" if c == "time_out" else c, fontsize=19)
        if j == 0:
            if g[0] is not None:
                ax.set_ylabel(g[1], fontsize=18)
            else:
                ax.set_ylabel(g[1], fontsize=18)

        # Remove chart borders and ticks
        ax.spines["top"].set_visible(False)
        ax.spines["right"].set_visible(False)
        ax.spines["bottom"].set_visible(False)
        ax.spines["left"].set_visible(False)
        ax.get_xaxis().set_ticks([])
        ax.get_yaxis().set_ticks([])

        # Get positions
        if g[0] is not None:
            row = generator.dataset.get_random_gen_class_rows(c, 1)
        else:
            row = dataset.get_random_class_rows(c, 1)

        xpoints = row.filter(regex="^#x").values.flatten()
        ypoints = row.filter(regex="^#y").values.flatten()
        vxpoints = row.filter(regex="^#vx").values.flatten()
        vypoints = row.filter(regex="^#vy").values.flatten()
        ballx = row.filter(regex="^#ball_x").values.flatten()
        bally = row.filter(regex="^#ball_y").values.flatten()

        tolerance = 0.05
        xpoints[np.isclose(xpoints, 0, atol=tolerance)] = 0
        ypoints[np.isclose(ypoints, 0, atol=tolerance)] = 0
        vxpoints[np.isclose(vxpoints, 0, atol=tolerance)] = 0
        vypoints[np.isclose(vypoints, 0, atol=tolerance)] = 0
        ballx[np.isclose(ballx, 0, atol=tolerance)] = 0
        bally[np.isclose(bally, 0, atol=tolerance)] = 0
        ids = ~(
            np.array(xpoints == 0)
            & np.array(ypoints == 0)
            & np.array(vxpoints == 0)
            & np.array(vxpoints == 0)
        )
        xpoints = xpoints[ids]
        ypoints = ypoints[ids]
        vxpoints = vxpoints[ids]
        vypoints = vypoints[ids]
        ids = ~(np.array(ballx == 0) & np.array(bally == 0))
        ballx = ballx[ids]
        bally = bally[ids]

        # First array horiz. coords., second vertical
        # Middle line
        ax.plot([0.5, 0.5], [0.0, 1.0], color="black")
        # Center circle
        ax.add_patch(Ellipse((0.5, 0.5), 0.2, 0.7, facecolor="none", ec="k", lw=2))
        # Areas
        ax.add_patch(
            Arc(
                (0.0, 0.5),
                0.2,
                0.9,
                angle=180,
                theta1=90,
                theta2=270,
                facecolor="none",
                ec="k",
                lw=2,
            )
        )
        ax.add_patch(
            Arc(
                (1.0, 0.5),
                0.2,
                0.9,
                theta1=90,
                theta2=270,
                facecolor="none",
                ec="k",
                lw=2,
            )
        )
        # Player positions
        ax.plot(xpoints, ypoints, "o")
        # Ball
        ax.plot(ballx, bally, "o", color="magenta")
        # Speeds
        ax.quiver(
            xpoints,
            ypoints,
            vxpoints,
            vypoints,
            color="orange",
            angles="uv",
            scale=4,
            headaxislength=3,
            headlength=3,
        )
        # Field
        ims.append(
            ax.add_patch(Rectangle((0, 0), 1, 1, facecolor="none", ec="k", lw=2))
        )

        axs.append(ax)

        j += 1

    i += 1

plt.savefig("figures/VisualPlaynet.pdf", format="pdf", bbox_inches="tight")



from . import Generator
from gentab.utils import console, DEVICE

import pandas as pd
from sdv.metadata import SingleTableMetadata
from sdv.single_table import CopulaGANSynthesizer
from sdv.sampling import Condition
import pandas as pd
from collections import Counter


class CopulaGAN(Generator):
    def __init__(
        self,
        dataset,
        enforce_min_max_values=True,
        enforce_rounding=False,
        epochs=300,
        batch_size=8000,
        discriminator_dim=(256, 256),
        discriminator_decay=1e-6,
        discriminator_lr=2e-4,
        discriminator_steps=1,
        embedding_dim=128,
        generator_decay=1e-6,
        generator_dim=(256, 256),
        generator_lr=2e-4,
        # https://github.com/sdv-dev/SDV/issues/1231 batch_size needs to be multiple of pac
        pac=10,
        cuda=True if DEVICE == "cuda" else False,
        locales=["en_US"],
        numerical_distributions=None,
        default_distribution="beta",
        log_frequency=True,
        max_tries_per_batch=4096,
    ) -> None:
        super().__init__(dataset)
        self.enforce_min_max_values = enforce_min_max_values
        self.enforce_rounding = enforce_rounding
        self.epochs = epochs
        self.batch_size = batch_size
        self.generator_dim = generator_dim
        self.discriminator_dim = discriminator_dim
        self.discriminator_decay = discriminator_decay
        self.discriminator_lr = discriminator_lr
        self.discriminator_steps = discriminator_steps
        self.embedding_dim = embedding_dim
        self.generator_decay = generator_decay
        self.generator_lr = generator_lr
        self.pac = pac
        self.cuda = cuda
        self.locales = locales
        self.numerical_distributions = numerical_distributions
        self.default_distribution = default_distribution
        self.log_frequency = log_frequency
        self.max_tries_per_batch = max_tries_per_batch

    def sample(self) -> pd.DataFrame:
        return super().sample()

    def preprocess(self) -> None:
        self.data = self.dataset.get_single_df()

    def train(self) -> None:
        metadata = SingleTableMetadata()
        metadata.detect_from_dataframe(data=self.data)

        self.synthesizer = CopulaGANSynthesizer(
            metadata,  # required
            enforce_min_max_values=self.enforce_min_max_values,
            enforce_rounding=self.enforce_rounding,
            default_distribution=self.default_distribution,
            epochs=self.epochs,
            batch_size=self.batch_size,
            generator_dim=self.generator_dim,
            discriminator_dim=self.discriminator_dim,
            embedding_dim=self.embedding_dim,
            discriminator_decay=self.discriminator_decay,
            discriminator_lr=self.discriminator_lr,
            discriminator_steps=self.discriminator_steps,
            generator_decay=self.generator_decay,
            generator_lr=self.generator_lr,
            locales=self.locales,
            numerical_distributions=self.numerical_distributions,
            pac=self.pac,
            log_frequency=self.log_frequency,
            cuda=self.cuda,
        )
        self.synthesizer.fit(self.data)

    def resample(self, n_samples, append) -> None:
        conditions = []
        for cls, cnt in n_samples.items():
            conditions.append(
                Condition(
                    num_rows=cnt, column_values={self.dataset.config["y_label"]: cls}
                )
            )

        data_gen = self.synthesizer.sample_from_conditions(
            conditions=conditions,
            batch_size=self.batch_size,
            max_tries_per_batch=self.max_tries_per_batch,
        )

        if append:
            self.dataset.set_split_result(
                pd.concat([self.data, data_gen], ignore_index=True, sort=False)
            )
        else:
            self.dataset.set_split_result(data_gen)

    def balance(self) -> None:
        conditions = []
        for cls, cnt in self.counts.items():
            conditions.append(
                Condition(
                    num_rows=cnt, column_values={self.dataset.config["y_label"]: cls}
                )
            )

        data_gen = self.synthesizer.sample_from_conditions(
            conditions=conditions,
            batch_size=self.batch_size,
            max_tries_per_batch=self.max_tries_per_batch,
        )

        self.dataset.set_split_result(
            pd.concat([self.data, data_gen], ignore_index=True, sort=False)
        )



from . import Generator
from gentab.utils import console, SPINNER, REFRESH

import pandas as pd
from imblearn.over_sampling import ADASYN as ada
from collections import Counter


class ADASYN(Generator):
    def __init__(self, dataset, n_neighbors=5, sampling_strategy="minority") -> None:
        super().__init__(dataset)
        self.n_neighbors = n_neighbors
        self.sampling_strategy = sampling_strategy
        self.X_ada = None

    def preprocess(self) -> None:
        self.X_ada = self.dataset.encode_categories(self.dataset.X)

    def train(self) -> None:
        super().train()

    def sample(self) -> pd.DataFrame:
        return super().sample()

    def resample(self, n_samples, append) -> None:
        if append:
            for cls, cnt in n_samples.items():
                n_samples[cls] += self.orig_counts[cls]

        self.dataset.X_gen, self.dataset.y_gen = ada(
            random_state=self.seed,
            sampling_strategy=n_samples,
            n_neighbors=self.n_neighbors,
        ).fit_resample(self.X_ada, self.dataset.y)

        self.dataset.X_gen = self.dataset.decode_categories(self.dataset.X_gen)

    def balance(self) -> None:
        if self.sampling_strategy == "minority":
            for _ in range(self.dataset.num_classes() - 1):
                self.dataset.X_gen, self.dataset.y_gen = ada(
                    random_state=self.seed,
                    sampling_strategy=self.sampling_strategy,
                    n_neighbors=self.n_neighbors,
                ).fit_resample(self.X_ada, self.dataset.y)
        else:
            self.dataset.X_gen, self.dataset.y_gen = ada(
                random_state=self.seed,
                sampling_strategy=self.sampling_strategy,
                n_neighbors=self.n_neighbors,
            ).fit_resample(self.X_ada, self.dataset.y)

        self.dataset.X_gen = self.dataset.decode_categories(self.dataset.X_gen)



import torch
import torch.nn as nn
import torch.nn.functional as F
from torch.optim import Adam
from .process_GQ import DataFrameParser
import gc
import random
import pandas as pd
import matplotlib.pyplot as plt


def _make_mlp_layers(num_units):
    layers = nn.ModuleList([
        nn.Linear(in_features, out_features)
        for in_features, out_features in zip(num_units, num_units[1:])
    ])
    return layers


class DeapStack(nn.Module):
    """Simple MLP body."""

    def __init__(
        self,
        n_bins,
        n_cats,
        n_nums,
        cards,
        in_features,
        hidden_size,
        bottleneck_size,
        num_layers,
    ):
        super().__init__()
        encoder_layers = num_layers >> 1
        decoder_layers = num_layers - encoder_layers - 1
        self.encoders = _make_mlp_layers([in_features] + [hidden_size] * encoder_layers)
        self.bottleneck = nn.Linear(hidden_size, bottleneck_size)
        self.decoders = _make_mlp_layers(
            [bottleneck_size] + [hidden_size] * decoder_layers
        )

        self.n_bins = n_bins
        self.n_cats = n_cats
        self.n_nums = n_nums

        self.bins_linear = nn.Linear(hidden_size, n_bins) if n_bins else None
        self.cats_linears = nn.ModuleList(
            [nn.Linear(hidden_size, card) for card in cards]
        )
        self.nums_linear = nn.Linear(hidden_size, n_nums) if n_nums else None

    def forward_pass(self, x):
        for encoder_layer in self.encoders:
            x = F.relu(encoder_layer(x))
        x = b = self.bottleneck(x)
        for decoder_layer in self.decoders:
            x = F.relu(decoder_layer(x))
        return [b, x]

    def forward(self, x):
        outputs = dict()

        num_min_values, _ = torch.min(
            x[:, self.n_bins + self.n_cats : self.n_bins + self.n_cats + self.n_nums],
            dim=0,
        )
        num_max_values, _ = torch.max(
            x[:, self.n_bins + self.n_cats : self.n_bins + self.n_cats + self.n_nums],
            dim=0,
        )

        decoder_output = self.forward_pass(x)[1]

        if self.bins_linear:
            outputs["bins"] = self.bins_linear(decoder_output)

        if self.cats_linears:
            outputs["cats"] = [linear(decoder_output) for linear in self.cats_linears]

        if self.nums_linear:
            before_threshold = self.nums_linear(decoder_output)
            outputs["nums"] = before_threshold

            for col in range(len(num_min_values)):
                outputs["nums"][:, col] = torch.where(
                    before_threshold[:, col] < num_min_values[col],
                    num_min_values[col],
                    before_threshold[:, col],
                )
                outputs["nums"][:, col] = torch.where(
                    before_threshold[:, col] > num_max_values[col],
                    num_max_values[col],
                    before_threshold[:, col],
                )

        return outputs

    def featurize(self, x):
        return self.forward_pass(x)[0]

    def decoder(self, latent_feature, num_min_values, num_max_values):
        decoded_outputs = dict()

        for layer in self.decoders:
            x = F.relu(layer(latent_feature))
        last_hidden_layer = x

        if self.bins_linear:
            decoded_outputs["bins"] = self.bins_linear(last_hidden_layer)

        if self.cats_linears:
            decoded_outputs["cats"] = [
                linear(last_hidden_layer) for linear in self.cats_linears
            ]

        if self.nums_linear:
            d_before_threshold = self.nums_linear(last_hidden_layer)
            decoded_outputs["nums"] = d_before_threshold

            for col in range(len(num_min_values)):
                decoded_outputs["nums"][:, col] = torch.where(
                    d_before_threshold[:, col] < num_min_values[col],
                    num_min_values[col],
                    d_before_threshold[:, col],
                )
                decoded_outputs["nums"][:, col] = torch.where(
                    d_before_threshold[:, col] > num_max_values[col],
                    num_max_values[col],
                    d_before_threshold[:, col],
                )

        return decoded_outputs


def auto_loss(inputs, reconstruction, n_bins, n_nums, n_cats, cards):
    """Calculating the loss for DAE network.
    BCE for masks and reconstruction of binary inputs.
    CE for categoricals.
    MSE for numericals.
    reconstruction loss is weighted average of mean reduction of loss per datatype.
    mask loss is mean reduced.
    final loss is weighted sum of reconstruction loss and mask loss.
    """
    bins = inputs[:, 0:n_bins]
    cats = inputs[:, n_bins : n_bins + n_cats]
    nums = inputs[:, n_bins + n_cats : n_bins + n_cats + n_nums]

    reconstruction_losses = dict()

    if "bins" in reconstruction:
        reconstruction_losses["bins"] = F.binary_cross_entropy_with_logits(
            reconstruction["bins"], bins
        )

    if "cats" in reconstruction:
        cats_losses = []
        for i in range(len(reconstruction["cats"])):
            cats_losses.append(
                F.cross_entropy(reconstruction["cats"][i], cats[:, i].long())
            )
        reconstruction_losses["cats"] = torch.stack(cats_losses).mean()

    if "nums" in reconstruction:
        reconstruction_losses["nums"] = F.mse_loss(reconstruction["nums"], nums)

    reconstruction_loss = torch.stack(list(reconstruction_losses.values())).mean()
    return reconstruction_loss


def sigmoid_threshold(logits):
    sigmoid_output = torch.sigmoid(logits)
    threshold_output = torch.where(
        sigmoid_output > 0.5, torch.tensor(1), torch.tensor(0)
    )
    return threshold_output


def softmax_with_max(predictions):
    # Applying softmax function
    probabilities = F.softmax(predictions, dim=1)

    # Getting the index of the maximum element
    max_indices = torch.argmax(probabilities, dim=1)

    return max_indices


def train_autoencoder(
    df,
    hidden_size,
    num_layers,
    lr,
    weight_decay,
    n_epochs,
    batch_size,
    threshold,
    progress,
    task,
):
    parser = DataFrameParser().fit(df, threshold)
    data = parser.transform()
    data = torch.tensor(data.astype("float32"))

    datatype_info = parser.datatype_info()
    n_bins = datatype_info["n_bins"]
    n_cats = datatype_info["n_cats"]
    n_nums = datatype_info["n_nums"]
    cards = datatype_info["cards"]

    DS = DeapStack(
        n_bins,
        n_cats,
        n_nums,
        cards,
        data.shape[1],
        hidden_size=128,
        bottleneck_size=data.shape[1],
        num_layers=3,
    )

    optimizer = Adam(DS.parameters(), lr=lr, weight_decay=weight_decay)

    # tqdm_epoch = tqdm.trange(n_epochs)

    losses = []
    batch_size
    all_indices = list(range(data.shape[0]))

    for epoch in range(n_epochs):
        batch_indices = random.sample(all_indices, batch_size)

        output = DS(data[batch_indices, :])

        l2_loss = auto_loss(
            data[batch_indices, :], output, n_bins, n_nums, n_cats, cards
        )
        optimizer.zero_grad()
        l2_loss.backward()
        optimizer.step()

        gc.collect()
        torch.cuda.empty_cache()

        # Print the training loss over the epoch.
        losses.append(l2_loss.item())

        # tqdm_epoch.set_description("Average Loss: {:5f}".format(l2_loss.item()))
        progress.update(task, completed=epoch + 1)

    num_min_values, _ = torch.min(
        data[:, n_bins + n_cats : n_bins + n_cats + n_nums], dim=0
    )
    num_max_values, _ = torch.max(
        data[:, n_bins + n_cats : n_bins + n_cats + n_nums], dim=0
    )

    latent_features = DS.featurize(data)
    output = DS.decoder(latent_features, num_min_values, num_max_values)

    return (DS.decoder, latent_features, num_min_values, num_max_values)
